/*
** Module   : UastHintsWorker.java
** Abstract : provides a pattern engine service to lookup generic UAST hints
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050314   @20326 Created initial version that provides
**                           support for reading string, long, double
**                           and boolean values using a file-unique key.
** 002 GES 20050316   @20398 Converted this worker to use the visitAst()
**                           interface in PatternWorker instead of using
**                           a rule driven approach.  
** 003 ECF 20050526   @21290 Changes to support new expression engine
**                           implementation. Changed library registration
**                           mechanism based on superclass' modifications
**                           to support single library per pattern worker
**                           limit.
** 004 ECF 20050713   @21700 Added getTableHints user function. Retrieves
**                           hints, if any, for a specified table.
** 005 GES 20051129   @23548 Fixes to remove null pointer exceptions.
** 006 ECF 20061126   @31401 Added getExternalTableHints() method to
**                           HintsReader. This method is used to read a
**                           set of hints for an explicit, external AST.
**                           Specifically, it allows temp/work-tables to
**                           read table hints for permanent tables.
** 007 GES 20070718   @34595 Fix to remove null pointer exception.
** 008 GES 20090429   @42067 Match package and class name changes.
** 009 GES 20090514   @42234 Removed dead code.
** 010 ECF 20091006   @44186 Allow a .dict AST to load UAST hints associated
**                           with the 4GL program file. This allows us to
**                           maintain a single hints file per program, but to
** 011 LMR 20110106          Fixes to reflect renaming of load() method in 
**                           Configuration class to getSchemaConfig() and 
**                           removal of its 'key' parameter.
** 012 VMN 20140423          Added enhance schema name conversion support for hint "escape"
**                           attribute.
*/

package com.goldencode.p2j.uast;

import java.io.*;
import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.schema.*;

/**
 * Provides a pattern engine service to read string, long, double and boolean
 * values using a file-unique key.
 */
public class UastHintsWorker
extends AbstractPatternWorker
{
   /** Stores the current AST's hints object. */
   private UastHints hints = null;
   
   /**
    * Default constructor which calls the super-class constructor, registers
    * its libraries and initializes its instance members.
    */
   public UastHintsWorker()
   throws ConfigurationException,
          AstException
   {
      super();
      setLibrary(new HintsReader());
   }
      
   /**
    * Resets the current hints instance to <code>null</code> each time a new
    * AST is visited.  This enables the lazy loading of hints files only when
    * needed.
    *
    * @param   ast
    *          The root node of the source AST about to be processed by the
    *          pattern engine.
    */
   public void visitAst(Aast ast)
   {
      hints = null;
   }
   
   /**
    * Provides a service for reading generic named hints associated with the
    * current AST.  This class is a <code>PatternEngine</code> worker.
    */
   public class HintsReader
   {
      /**
       * Retrieve the table hints, if any, for the specified table name.
       *
       * @param   table
       *          Name of the table for which hint information should be
       *          retrieved.
       *
       * @return  Table hints instance for the specified table, or
       *          <code>null</code> if none exists.
       */
      public TableHints getTableHints(String table)
      {
         if (!loadHints())
         {
            return null;
         }
         
         Map map = hints.getTableHints();
         if (map == null)
         {
            return null;
         }
         
         return (TableHints) map.get(table);
      }
      
      /**
       * Retrieve the table hints, if any, for the specified table name from
       * a hints file not directly associated with the currently loaded AST.
       * Instead the hints file is assumed to be a database-related hints
       * file identified by the specified, fully qualified table name.
       * <p>
       * Note:  unlike the other service methods of this worker, the hints
       * loaded by this method are not cached, but are read each time this
       * method is invoked.
       *
       * @param   table
       *          Name of the table for which hint information should be
       *          retrieved.  This name is a schema name which must be
       *          qualified by the target database name for which to load
       *          hints.
       *
       * @return  Table hints instance for the specified table, or
       *          <code>null</code> if none exists.
       */
      public TableHints getExternalTableHints(String table)
      {
         // Load the external hints file, if any.
         EntityName ename = new EntityName(EntityName.TABLE, table);
         UastHints extHints = null;
         
         try
         {
            String database = ename.getDatabase();
            SchemaConfig config = Configuration.getSchemaConfig();
            File schemaFile = config.getSchemaFile(database);
            
            if (schemaFile == null)
            {
               return null;
            }
            
            String path = schemaFile.getAbsolutePath();
            int dotPos = path.lastIndexOf(".");
            StringBuilder buf = new StringBuilder(path.substring(0, dotPos));
            buf.append(".schema");
            extHints = new UastHints(buf.toString());
         }
         
         catch (Exception exc)
         {
            exc.printStackTrace();
            return null;
         }
         
         Map map = extHints.getTableHints();
         if (map == null)
         {
            return null;
         }
         
         return (TableHints) map.get(ename.getTable());
      }
      
      /**
       * Get a UAST hint using a uniquely named key, accessing a string or
       * <code>null</code> if no such hint exists.
       *
       * @param    key
       *           Unique name for the hint, which is used as a key.
       *
       * @return   The named string or <code>null</code>.
       */
      public String getUastString(String key)
      {
         if (!loadHints())
            return null;         
         
         String result = null;
         
         String cls = hints.getUastType(key);
         
         if (cls != null && cls.equals("java.lang.String"))
         {
            result = hints.getUastString(key);
         }
         
         return result;
      }      
         
      /**
       * Get a UAST hint using a uniquely named key, accessing a string array
       * or <code>null</code> if no such hint exists.
       *
       * @param    key
       *           Unique name for the hint, which is used as a key.
       * @param    index
       *           The 0-based index of the entry to access.
       *
       * @return   The indexed element in the named string array or 
       *           <code>null</code> if the array or element doesn't exist.
       */
      public String getUastStringArray(String key, int index)
      {
         if (!loadHints())
            return null;
         
         String[] array  = null;
         String   result = null;
         String   cls    = hints.getUastType(key);
         
         if (cls != null && cls.equals("[Ljava.lang.String;"))
         {
            array = hints.getUastStringArray(key);
         }
         
         try
         {
            if (array != null)
            {
               result = array[index];
            }
         }
         
         catch (IndexOutOfBoundsException ioe)
         {
            // nothing to do, just return null
         }
         
         return result;
      }      
         
      /**
       * Get a UAST hint using a uniquely named key, accessing a long or
       * <code>null</code> if no such hint exists.
       *
       * @param    key
       *           Unique name for the hint, which is used as a key.
       *
       * @return   The named long or <code>null</code>.
       */
      public Long getUastLong(String key)
      {
         if (!loadHints())
            return null;
         
         Long   result = null;
         String cls    = hints.getUastType(key);
         
         if (cls != null && cls.equals("java.lang.Long"))
         {
            result = hints.getUastLong(key);
         }
         
         return result;
      }      
               
      /**
       * Get a UAST hint using a uniquely named key, accessing a double or
       * <code>null</code> if no such hint exists.
       *
       * @param    key
       *           Unique name for the hint, which is used as a key.
       *
       * @return   The named double or <code>null</code>.
       */
      public Double getUastDouble(String key)
      {
         if (!loadHints())
            return null;
         
         Double result = null;
         String cls    = hints.getUastType(key);
         
         if (cls != null && cls.equals("java.lang.Double"))
         {
            result = hints.getUastDouble(key);
         }
         
         return result;
      }      
         
      /**
       * Get a UAST hint using a uniquely named key, accessing a boolean or
       * <code>null</code> if no such hint exists.
       *
       * @param    key
       *           Unique name for the hint, which is used as a key.
       *
       * @return   The named boolean or <code>null</code>.
       */
      public Boolean getUastBoolean(String key)
      {
         if (!loadHints())
            return null;
         
         Boolean result = null;
         String  cls    = hints.getUastType(key);
         
         if (cls != null && cls.equals("java.lang.Boolean"))
         {
            result = hints.getUastBoolean(key);
         }
         
         return result;
      }      
      
      /**
       * Accesses the class name (data type) of a UAST hint using a uniquely
       * named key.
       *
       * @param    key
       *           Unique name for the hint, which is used as a key.
       *
       * @return   The class name of the hint if it exists or 
       *           <code>null</code> if the hint does not exist.
       */
      public String getUastHintType(String key)
      {
         if (!loadHints())
            return null;
         
         return hints.getUastType(key);
      }      
      
      /**
       * Checks the existence of a UAST hint using a uniquely named key.
       *
       * @param    key
       *           Unique name for the hint, which is used as a key.
       *
       * @return   <code>true</code> if any hint exists by that key for this
       *           file, <code>false</code> if no such hint exists.
       */
      public boolean isUastHint(String key)
      {
         if (!loadHints())
            return false;
         
         return hints.isUastHint(key);
      }

      /**
       * Checks if hint is set for schema.
       *
       * @return   <code>true</code> if "escape" hint is set for schema and equals
       *           <code>true</code>, otherwise <code>false</code>.
       */
      public boolean isSchemaEscaped()
      {
         if (!loadHints())
            return false;

         return hints.isSchemaEscaped();
      }

      /**
       * Checks if field should be escaped.
       *
       * @param    table
       *           The legacy table name.
       * @param    field
       *           The legacy field name.
       *
       * @return   value of "escape" hint for field if hint is set, otherwise "escape" hint for
       *           table or schema.
       */
      public boolean isFieldEscaped(String table, String field)
      {
         if (!loadHints())
            return false;

         return hints.isFieldEscaped(table, field);
      }

      /**
       * Checks if table should be escaped.
       *
       * @param    table
       *           The legacy table name.
       *
       * @return   value of "escape" hint for table if hint is set, otherwise "escape" hint for
       *           schema.
       */
      public boolean isTableEscaped(String table)
      {
         if (!loadHints())
            return false;

         return hints.isTableEscaped(table);
      }

      /**
       * Loads the hints file for an AST if it exists and hasn't already been
       * loaded.
       *
       * @return   <code>true</code> if there are hints for this file and the
       *           hints are loaded, <code>false</code> if no hints are
       *           available.
       */
      private boolean loadHints()
      {
         if (hints == null)
         {
            try
            {
               Aast src  = resolver.getSourceAst();
               String filename = src.getFilename();
               if (filename.toLowerCase().endsWith(AstGenerator.DICT_POSTFIX))
               {
                  int len = AstGenerator.DICT_POSTFIX.length();
                  filename = filename.substring(0, filename.length() - len);
               }
               File file = Configuration.toFile(filename);
               
               hints = new UastHints(file.getAbsolutePath());
            }
            catch (Exception exc)
            {
               // any failure drops us out
               return false;
            }
         }
         
         return true;
      }
   }
}
